//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#ifndef BIHSAHTREEBUILDER_H_
#define BIHSAHTREEBUILDER_H_

#include "Types.h"
#include "Helpers.h"
#include <assert.h>


class BIHSAHTreeBuilder
{
	private:
		
		enum Type
		{
			Start,
			End
		};
		struct Event
		{
			int primitive;
			Type type;
			float position(float* min, float* max) const
			{
				if (type == Type::Start)
					return min[primitive];
				else
					return max[primitive];
			}
			Event(int primitive, Type type): primitive(primitive), type(type)
			{
			}
			Event()
			{
			}
		};
		class EventCompare {
			private:
				float** mins;
				float** maxs;
				int axis;

			public:
				EventCompare(float* mins[3], float* maxs[3], int axis):axis(axis)
				{
					this->mins = mins;
					this->maxs = maxs;
				}
				//Return true if s1 < s2; otherwise, return false.
				bool operator()(const Event e1, const Event e2) {
					float pos1 = e1.position(mins[axis], maxs[axis]);
					float pos2 = e2.position(mins[axis], maxs[axis]);
					float c = pos2-pos1;
					int add = 1;
					while (c == 0 && add < 3)
					{
						assert(pos1 == pos2);
						pos1 = e1.position(mins[modulo[axis+add]], maxs[modulo[axis+add]]);
						pos2 = e2.position(mins[modulo[axis+add]], maxs[modulo[axis+add]]);
						c = pos2-pos1;
						add++;
					}
					return c > 0;
				}
		};
		static BIHNode* BIHSAHTreeBuilder::BuildTreeRecursivly(Event* events[3], float* primitive_mins[3], float* primitive_maxs[3], int event_start, int event_end, Vector boundary_min, Vector boundary_max, int* result_indices);
		static int BIHPartition(Event* events, int left, int right, float pivotValue, float* primitive_maxs);
		static void BIHExchange(Event* events, int a, int b);

	public:

		static BIHTree* Compile(Scene * scene);
		 
};



#endif /* BIHSAHTREEBUILDER_H_ */