//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include <iostream>
#include <vector>
#include <queue>
#include <fstream>
#include <string>
#include <sstream>

#include <assert.h>
#include <math.h>

#include "Helpers.h"

using namespace std;

class delimiterstringstream
{
private:
	string s;
	string d;
	string::size_type start;
	string::size_type end;

	template<class T> static T fromString(const std::string& s){
		     std::istringstream stream (s);
		     T t;
		     stream >> t;
		     return t;
	}
	string currentSubstring()
	{
		return s.substr(start, end - start + 1);
	}

	int getNextDelimiter()
	{
		int pos = s.find_first_of(d, start);
		return pos == -1 ? s.length() : pos;
	}

public:
	delimiterstringstream(const string& s, const string& delimiters): s(s), d(delimiters) {
		start = 0;
	    end = getNextDelimiter()-1;
	}

	~delimiterstringstream(){
	}

	template<class T> delimiterstringstream& operator>>(T& t){

		if (start < s.length())
		{
			string current = currentSubstring();
			if (current.empty())
			{
				t = 0;
				start++;
				end = getNextDelimiter()-1;
			}
			else
			{
				t = fromString<T>(current);
				start = end + 1 + 1; //because we go over the delimiter
				end = getNextDelimiter()-1;
			}
			return *this;
		}
		else
		{
			throw "end of string!";
		}
	}

};

void Helpers::LoadSceneFromFile(Scene* scene, char *fileName)
{

	vector<Point> vertexVector;
	vector<Normal> normalVector;
	vector<int> vertexIndexVector;
	vector<int> normalIndexVector;

	string line;
	ifstream myfile(fileName);

	if (myfile.is_open()) {
		while (!myfile.eof()) {
			getline(myfile,line);

			string buf;
			stringstream ss(line);

			ss >> buf;

			if (buf == "v") {
				float x,y,z;
				ss >> x;
				ss >> y;
				ss >> z;
				vertexVector.push_back(Point(x,y,z));
			} else if ( buf == "vn" ) {
				float x,y,z;
				ss >> x;
				ss >> y;
				ss >> z;
				normalVector.push_back(Normal(x,y,z));
			} else if ( buf == "f" ) {
				// we iterate over the triangle corners
				for (int index = 0; index < 3; index++)
				{
					string corner;
					ss >> corner;
					delimiterstringstream ds(corner, "/");
					int vertexIndex, uvIndex, normalIndex;
					ds >> vertexIndex;
					ds >> uvIndex;
					ds >> normalIndex;
					if (vertexIndex > 0)
					{
						vertexIndexVector.push_back(vertexIndex);
					}
					if (normalIndex > 0)
					{
						normalIndexVector.push_back(normalIndex);
					}
				}
			}
		}
	}

	assert(vertexIndexVector.size() == normalIndexVector.size());
	std::size_t count = vertexIndexVector.size();

	/*scene->points = (Point*)malloc(count*sizeof(Point));
	scene->normals = (Point*)malloc(count*sizeof(Normal));*/
	
	scene->points = new Point[count];
	scene->normals = new Point[count];

	scene->count = count;

	Point* points = scene->points;
	Point* normals = scene->normals;

	for (std::size_t index = 0; index < count; index++)
	{
		points[index] = vertexVector[vertexIndexVector[index]-1];
		normals[index] = normalVector[normalIndexVector[index]-1].normalize();
	}

	/*std::copy(vertexVector.begin(), vertexVector.end(), scene->points);
	std::copy(vertexIndexVector.begin(), vertexIndexVector.end(), scene->indices);*/
}

unsigned int Helpers::LongestAxis(float length[3])
{
	unsigned int maxAxis;
	float maxLength = -1;
	for (unsigned int index = 0; index < 3; index++) {
		if (length[index] > maxLength) {
			maxAxis = index;
			maxLength = length[index];
		}
	}
	return maxAxis;
}

void Helpers::GetBoundingBox(Vector* points, int count, Vector& min, Vector& max)
{
	min[0] = MAX_DISTANCE;
	min[1] = MAX_DISTANCE;
	min[2] = MAX_DISTANCE;
	max[0] = -MAX_DISTANCE;
	max[1] = -MAX_DISTANCE;
	max[2] = -MAX_DISTANCE;

	for (int index = 0; index < count; index++)
	{
		Vector v = *points;
		for (int axis = 0; axis < 3; axis++)
		{
			if (v[axis] < min[axis])
				min[axis] = v[axis];
			if (v[axis] > max[axis])
				max[axis] = v[axis];
		}
		points++;
	}
}

bool Helpers::CheckBoundingBox(VertexTriangle* triangles, int start, int end, Vector min, Vector max)
{
	Vector* points = (Vector*)&triangles[start];
	int length = (end - start) * 3;
	Vector correctMin;
	Vector correctMax;
	GetBoundingBox(points, length, correctMin, correctMax);
	return (correctMin == min && correctMax == max);
}

int Helpers::CountItems(const BIHNode * node)
{
	if (node->isLeaf)
	{
		return node->leaf.count;
	}
	else
	{
		return CountItems(node->inner.childs[0]) + CountItems(node->inner.childs[1]);
	}
}

int Helpers::CountItems(const BIHNode2* node)
{
	if (ISLEAF(node->index))
	{
		return node->data.leaf.count;
	}
	else
	{
		return CountItems(node+(INDEX(node->index))) + CountItems(node+(INDEX(node->index))+1);
	}
}

int Helpers::CountNodes(BIHNode* node)
{
	if (node->isLeaf)
	{
		return 1;
	}
	else
	{
		return 1 + CountNodes(node->inner.childs[0]) + CountNodes(node->inner.childs[1]);
	}
}

BIHNode2* Helpers::BIHTransformTree(BIHNode* root)
{
	int count = Helpers::CountNodes(root);
	BIHNode2* newNodes = (BIHNode2*)malloc(count*sizeof(BIHNode2));
	//BIHNode2* newNodes = new BIHNode2[count];
	queue<BIHNode*> q;
	int mark = 0;
	q.push(root);
	// do a level order tree walk
	while (!q.empty())
	{
		BIHNode* node = q.front();
		q.pop();
		std::cout << "size is" << q.size() << "\n";
		if (node->isLeaf)
		{
			newNodes[mark++] = BIHNode2(node->leaf.item, node->leaf.count);
		}
		else
		{
			q.push(node->inner.childs[0]);
			int add = q.size();
			newNodes[mark++] = BIHNode2(node->inner.axis, add, node->inner.boundaries);
			q.push(node->inner.childs[1]);
		}
	}
	assert(mark == count);
	return newNodes;
}


TriAccel* Helpers::BuildTriAccels(VertexTriangle* triangles, NormalTriangle* normals, int count)
{
	TriAccel* triAccels = new TriAccel[count];
	for (int index = 0; index < count; index++)
	{
		TriAccel& triAccel = triAccels[index];
		VertexTriangle& tri = triangles[index];
		Normal n = (tri.b-tri.a).cross(tri.c-tri.b);
		Normal averaged_normal = (normals[index].a+normals[index].b+normals[index].c).normalize();
		if (n*averaged_normal < 0)
			n = -n;
		int k;
		if ( fabs(n.x) > fabs(n.y) )
			if ( fabs(n.x) > fabs(n.z) )
				k = 0;
			else
				k = 2;
		else
			if ( fabs(n.y) > fabs(n.z) )
				k = 1;
			else
				k = 2;
		int u = modulo[k+1];
		int v = modulo[k+2];

		triAccel.k = k;
		triAccel.n_u = n[u]/n[k];
		triAccel.n_v = n[v]/n[k];
		
		triAccel.n_d = tri.a[k]+tri.a[u]*triAccel.n_u+tri.a[v]*triAccel.n_v;

		Vector b = tri.a-tri.b;
		Vector c = tri.c-tri.b;


		float div = b[u]*c[v]-b[v]*c[u];
		assert(div != 0);

		//beta = (b[u]*h[v] - b[v]*h[u])/div;
		triAccel.b_nu = -b[v]/div;
		triAccel.b_nv = b[u]/div;
		triAccel.b_d = (b[v]*tri.b[u]-b[u]*tri.b[v])/div;
		
		//gamma = (h[u]*c[v] - h[v]*c[u])/div;
		triAccel.c_nu = c[v]/div;
		triAccel.c_nv = -c[u]/div;
		triAccel.c_d = (-c[v]*tri.b[u]+c[u]*tri.b[v])/div;		
	}
	return triAccels;
}

int Helpers::GetDepth(BIHNode* node)
{
	if (node->isLeaf)
		return 1;
	else
		return 1+max(GetDepth(node->inner.childs[0]), GetDepth(node->inner.childs[1]));
}






