//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#ifndef RAYTRACER_H_
#define RAYTRACER_H_

#include <windows.h>		// Header File For Windows
#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <gl\glu.h>			// Header File For The GLu32 Library

#include "Types.h"

//#define SCREEN_SIZE 512
#define BLOCK_SIZE 16
#define NUMBER_OF_THREADS 16

struct Intersection
{
	static Intersection NO_INTERSECTION;
	float t;
	float beta;
	float gamma;
	unsigned int index;
	bool operator==(const Intersection& i);
	bool operator<(const Intersection& i);
	bool operator>(const Intersection& i);
	Intersection(float t, unsigned int index, float beta, float gamma);
	Intersection();
	Intersection(const Intersection& i);
};

class Raytracer {

private:
	// calculates the direct lighting component
	static Color RaytraceLight(Scene *scene, BIHTree2* tree, const Light& light, const Vector& surfacePosition, const Normal& surfaceNormal);
	// shades the given surface point
	static Color Shade(const Intersection& i, Scene* scene, BIHTree2* tree, const Ray& ray, Normal& normal);
	// raytrace a single ray recursion
	static void RaytraceRecursivly(const BIHNode2* node, const Ray& ray, TriAccel* triangles, const float tmin, const float tmax, Intersection& result);
	// shadow tracing recursion
	static bool BoolRaytrace(const BIHNode2* node, const Ray& ray, TriAccel* triangles, const float tmin, const float tmax);

public:
	// intersection with aabb
	static bool Intersect(const Ray & r, const Vector bounds[2], float tmin, float tmax, float r_interval[2]);
	// intersection with triangle
	static bool Intersect(const Ray & r, const Point & A, const Point & B, const Point & C, const Normal & n, float& t, float& beta, float& gamma);
	// intersection with acceleration struct for triangle
	static bool Intersect(const Ray& r, const TriAccel& acc, float& t, float& beta, float& gamma, float tMin, float tMax);

	// raytrace the scene for the given screen
	static void Raytrace(Scene *scene, BIHTree2* tree, const Transform & transform, GLubyte *image, GLsizei imageWidth, GLsizei imageHeight, GLsizei pixel_size);
	// raytracing the tree for shadows
	static bool BoolRaytrace(Scene *scene, BIHTree2* tree, const Ray& ray);
	// raytraces a single ray
	static bool Raytrace(Scene* scene, BIHTree2* tree, const Ray& ray, Intersection& i);
	
};

#endif /* RAYTRACER_H_ */
