//=====================================================================================================================
//
// Unnamed Realtime Ray Tracer - Copyright by Jyrki Friedrich. All rights reserved.
// You are free to use Unnamed Realtime Ray Tracer for non-commercial purposes.
// for all other projects: mail t4rj4n@gmail.com
//
//=====================================================================================================================

#include <windows.h>		// Header File For Windows
#include <Windowsx.h>
#include <gl\gl.h>			// Header File For The OpenGL32 Library
#include <gl\glu.h>			// Header File For The GLu32 Library
#include <sstream>
#include <omp.h>

#define _USE_MATH_DEFINES
#include <math.h>

#include "glfont2.h"
#include "hr_time.h"
#include <assert.h>

#include "Types.h"
#include "Raytracer.h"
#include "Helpers.h"
#include "Settings.h"
#include "BIHRandTreeBuilder.h"
#include "BIHSAHTreeBuilder.h"
#include "Tests.h"

#define VK_W 0x57
#define VK_A 0x41
#define VK_S 0x53
#define VK_D 0x44
#define VK_E 0x45
#define VK_C 0x43
#define VK_F1 0x70
#define VK_F2 0x71
#define VK_F3 0x72
#define VK_F4 0x73
#define VK_F5 0x74
#define VK_F6 0x75

#define MOVEMENT_SPEED 0.1f
#define ROTATION_SPEED 0.5f

using namespace glfont;

#include<iostream>
using namespace std;


HDC			hDC=NULL;		// Private GDI Device Context
HGLRC		hRC=NULL;		// Permanent Rendering Context
HWND		hWnd=NULL;		// Holds Our Window Handle
HINSTANCE	hInstance;		// Holds The Instance Of The Application

bool	keys[256];			// Array Used For The Keyboard Routine
bool	active=TRUE;		// Window Active Flag Set To TRUE By Default
bool	fullscreen=TRUE;	// Fullscreen Flag Set To Fullscreen Mode By Default

LRESULT	CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);	// Declaration For WndProc

GLubyte *image;

GLsizei imageWidth;
GLsizei imageHeight;

GLFont myfont;
CStopWatch stateStopWatch;
CStopWatch renderStopWatch;

//double lastElapsedTime = 0;

Scene* scene;
BIHTree2* tree2;
float speed;

UINT INPUT_TIMER_ID;

Camera camera(Vector(0,0,0),0,0);
Transform transformData;
const string info = "UnRR, v. 0.2";

GLvoid ReSizeGLScene(GLsizei width, GLsizei height)		// Resize And Initialize The GL Window
{
	if (image == NULL)
		;
	else {
		delete image;
	}
	image = new GLubyte[width*height*sizeof(GLubyte)*4];

	imageWidth = width;
	imageHeight = height;

	if (height==0)										// Prevent A Divide By Zero By
	{
		height=1;										// Making Height Equal One
	}

	//Initialize the viewport
	glViewport(0, 0, width, height);

	//Initialize OpenGL projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0.0, width, 0.0, height);

}


void RaytraceScene(const Transform & transform) {
	Raytracer::Raytrace(scene, tree2, transform, image, imageWidth, imageHeight, sizeof(GLubyte)*4);
}

void CreateScene(char* fileName = "..\\scenes\\bunny.obj") {
	scene = new Scene();
	Helpers::LoadSceneFromFile(scene, fileName);
	scene->lights.push_back(Light(Vector(1,-1,-1), DirectionalLight, Color(1.0f,0.95f,0.8f)));
	/*scene->lights.push_back(Light(Vector(1,-1,1), DirectionalLight, Color(1.0f,0.7f,0.7f)));
	scene->lights.push_back(Light(Vector(0,1,0), DirectionalLight, Color(0.7f,1.0f,0.7f)));
	scene->lights.push_back(Light(Vector(0,0,0), AmbientLight, Color(0.1f,0.1f,0.1f)));*/
	//scene->lights.push_back(Light(Vector(-2,1,-1), DirectionalLight, Color(1.0f,0.0f,0.0f)));
	//scene->lights.push_back(Light(Vector(0,1,1), DirectionalLight, Color(0.0f,0.0f,1.0f)));
	//scene->lights.push_back(Light(Vector(-1,1,1), DirectionalLight, Color(1,0,0)));
	//BIHTree* tree = BIHRandTreeBuilder::Compile(scene);
	
	BIHTree* tree = BIHSAHTreeBuilder::Compile(scene);
	tree2 = new BIHTree2(Helpers::BIHTransformTree(tree->root), tree->min, tree->max, Helpers::GetDepth(tree->root), Helpers::CountNodes(tree->root));
	assert(Helpers::CountItems(tree->root) == Helpers::CountItems(tree2->root));
	assert(Helpers::CountItems(tree->root) == (scene->count)/3);
	delete tree;
	scene->triAccels = Helpers::BuildTriAccels((VertexTriangle*)scene->points, (NormalTriangle*)scene->normals, (scene->count)/3);
	Vector position(tree2->min[0]+(tree2->max[0]-tree2->min[0])/2, tree2->min[1]+(tree2->max[1]-tree2->min[1])/2, tree2->max[2]+(tree2->max[2]-tree2->min[2])/2);
	camera.position = position;
	speed = (tree2->max - tree2->min).length()*MOVEMENT_SPEED;
}

string doubleToString(double d){
// Objekt auf die Klasse stringstream erstellen.
	stringstream ss; 
	ss << d;
	return (ss.str());
}

string intToString(int i){
// Objekt auf die Klasse stringstream erstellen.
	stringstream ss; 
	ss << i;
	return (ss.str());
}

void HandleInput()
{
	
	/*if (capturing)
	{
		camera.horizontalAngle += (float)mouseDelta.x * ROTATION_SPEED * lastElapsedTime;
		camera.verticalAngle += (float)mouseDelta.y * ROTATION_SPEED * lastElapsedTime;
		mouseDelta.x = 0;
		mouseDelta.y = 0;
	}*/
	
	stateStopWatch.stopTimer();
	double lastElapsedTime = stateStopWatch.getElapsedTime();
	stateStopWatch.startTimer();
	int forward = keys[VK_W] ? 1 : -keys[VK_S];
	int leftright = keys[VK_A] ? -!keys[VK_D] : keys[VK_D];
	int updown = keys[VK_E] ? -!keys[VK_C] : keys[VK_C];
	camera.horizontalAngle -= (float)leftright * ROTATION_SPEED * lastElapsedTime;
	camera.verticalAngle -= (float)updown * ROTATION_SPEED * lastElapsedTime;
	Vector movement(0, 0, 2*(-forward));
	Matrix m = Matrix::rotationY(camera.horizontalAngle);
	Matrix m2 = Matrix::rotationX(camera.verticalAngle);

	Matrix im = Matrix::rotationY(-camera.horizontalAngle);
	Matrix im2 = Matrix::rotationX(-camera.verticalAngle);

	Matrix rotationMatrix = m*m2;
	movement = rotationMatrix * movement;
	movement = movement * speed * lastElapsedTime;
	camera.position = camera.position + movement;
	
	transformData = Transform(rotationMatrix, im2*im, camera.position);
}

DWORD WINAPI MyTimer(LPVOID lpParam) 
{ 
	while(true) {
		HandleInput();
		Sleep(100);
	}
	return 0;
} 

int InitGL(GLvoid)										// All Setup For OpenGL Goes Here
{
	//Initialize OpenGL
	glClearColor(0.0, 0.0, 0.0, 1.0);	
	glEnable(GL_TEXTURE_2D);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//Create our glFont from verdana.glf, using texture 1
	if (!myfont.Create("font.glf", 1))
		return FALSE;

	stateStopWatch.startTimer();
	renderStopWatch.startTimer();
	omp_set_num_threads(NUMBER_OF_THREADS);
	HANDLE hThread = CreateThread(NULL, 0, MyTimer, NULL, 0, NULL);


	return TRUE;										// Initialization Went OK
}

int DrawGLScene(GLvoid)									// Here's Where We Do All The Drawing
{
	renderStopWatch.stopTimer();
	double lastElapsedTime = renderStopWatch.getElapsedTime();
	renderStopWatch.startTimer();

	HandleInput();
	Transform t = transformData;
	RaytraceScene(t);
	//Initialize modelview matrix
	glMatrixMode(GL_MODELVIEW);
	
	glClear(GL_COLOR_BUFFER_BIT); 						// Clear Screen Buffer
	glLoadIdentity();									// Reset The Current Modelview Matrix
	glDrawPixels(imageWidth, imageHeight, GL_RGBA, GL_UNSIGNED_BYTE, image);
	
	//Draw some strings
	glColor3f(1.0F, 1.0F, 1.0F);
	string s2 = doubleToString(1.0/lastElapsedTime)+" fps";
	//string s3 = doubleToString(((double)imageWidth*imageHeight)/lastElapsedTime)+ " rays/s";
	string s = "scene with "+intToString(scene->count)+" tris";
	myfont.Begin();
	myfont.DrawString(info, 10, imageHeight-5);
	std::pair<int,int> size;
	myfont.GetStringSize(s, &size);
	int posY = imageHeight-5;
	myfont.DrawString(s, imageWidth-size.first-10, posY);
	posY -= size.second - 5;
	myfont.GetStringSize(s2, &size);
	myfont.DrawString(s2, imageWidth-size.first-10, posY);
	/*posY -= size.second - 5;
	myfont.GetStringSize(s3, &size);
	myfont.DrawString(s3, imageWidth-size.first-10, posY);*/
	
	return TRUE;										// Keep Going
}

GLvoid KillGLWindow(GLvoid)								// Properly Kill The Window
{
	
	if (fullscreen)										// Are We In Fullscreen Mode?
	{
		ChangeDisplaySettings(NULL,0);					// If So Switch Back To The Desktop
		ShowCursor(TRUE);								// Show Mouse Pointer
	}

	if (hRC)											// Do We Have A Rendering Context?
	{
		if (!wglMakeCurrent(NULL,NULL))					// Are We Able To Release The DC And RC Contexts?
		{
			MessageBox(NULL,"Release Of DC And RC Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}

		if (!wglDeleteContext(hRC))						// Are We Able To Delete The RC?
		{
			MessageBox(NULL,"Release Rendering Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		}
		hRC=NULL;										// Set RC To NULL
	}

	if (hDC && !ReleaseDC(hWnd,hDC))					// Are We Able To Release The DC
	{
		MessageBox(NULL,"Release Device Context Failed.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hDC=NULL;										// Set DC To NULL
	}

	/*if (hWnd && !KillTimer(hWnd, INPUT_TIMER_ID))
	{
		MessageBox(NULL,"Could Not Release Timer.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
	}*/


	if (hWnd && !DestroyWindow(hWnd))					// Are We Able To Destroy The Window?
	{
		MessageBox(NULL,"Could Not Release hWnd.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hWnd=NULL;										// Set hWnd To NULL
	}


	if (!UnregisterClass("OpenGL",hInstance))			// Are We Able To Unregister Class
	{
		MessageBox(NULL,"Could Not Unregister Class.","SHUTDOWN ERROR",MB_OK | MB_ICONINFORMATION);
		hInstance=NULL;									// Set hInstance To NULL
	}
}

/*	This Code Creates Our OpenGL Window.  Parameters Are:					*
 *	title			- Title To Appear At The Top Of The Window				*
 *	width			- Width Of The GL Window Or Fullscreen Mode				*
 *	height			- Height Of The GL Window Or Fullscreen Mode			*
 *	bits			- Number Of Bits To Use For Color (8/16/24/32)			*
 *	fullscreenflag	- Use Fullscreen Mode (TRUE) Or Windowed Mode (FALSE)	*/

BOOL CreateGLWindow(char* title, int width, int height, int bits, bool fullscreenflag)
{
	GLuint		PixelFormat;			// Holds The Results After Searching For A Match
	WNDCLASS	wc;						// Windows Class Structure
	DWORD		dwExStyle;				// Window Extended Style
	DWORD		dwStyle;				// Window Style
	RECT		WindowRect;				// Grabs Rectangle Upper Left / Lower Right Values
	WindowRect.left=(long)0;			// Set Left Value To 0
	WindowRect.right=(long)width;		// Set Right Value To Requested Width
	WindowRect.top=(long)0;				// Set Top Value To 0
	WindowRect.bottom=(long)height;		// Set Bottom Value To Requested Height

	fullscreen=fullscreenflag;			// Set The Global Fullscreen Flag

	hInstance			= GetModuleHandle(NULL);				// Grab An Instance For Our Window
	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw On Size, And Own DC For Window.
	wc.lpfnWndProc		= (WNDPROC) WndProc;					// WndProc Handles Messages
	wc.cbClsExtra		= 0;									// No Extra Window Data
	wc.cbWndExtra		= 0;									// No Extra Window Data
	wc.hInstance		= hInstance;							// Set The Instance
	wc.hIcon			= LoadIcon(NULL, IDI_WINLOGO);			// Load The Default Icon
	wc.hCursor			= LoadCursor(NULL, IDC_ARROW);			// Load The Arrow Pointer
	wc.hbrBackground	= NULL;									// No Background Required For GL
	wc.lpszMenuName		= NULL;									// We Don't Want A Menu
	wc.lpszClassName	= "OpenGL";								// Set The Class Name

	if (!RegisterClass(&wc))									// Attempt To Register The Window Class
	{
		MessageBox(NULL,"Failed To Register The Window Class.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;											// Return FALSE
	}

	if (fullscreen)												// Attempt Fullscreen Mode?
	{
		DEVMODE dmScreenSettings;								// Device Mode
		memset(&dmScreenSettings,0,sizeof(dmScreenSettings));	// Makes Sure Memory's Cleared
		dmScreenSettings.dmSize=sizeof(dmScreenSettings);		// Size Of The Devmode Structure
		dmScreenSettings.dmPelsWidth	= width;				// Selected Screen Width
		dmScreenSettings.dmPelsHeight	= height;				// Selected Screen Height
		dmScreenSettings.dmBitsPerPel	= bits;					// Selected Bits Per Pixel
		dmScreenSettings.dmFields=DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

		// Try To Set Selected Mode And Get Results.  NOTE: CDS_FULLSCREEN Gets Rid Of Start Bar.
		if (ChangeDisplaySettings(&dmScreenSettings,CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL)
		{
			// If The Mode Fails, Offer Two Options.  Quit Or Use Windowed Mode.
			if (MessageBox(NULL,"The Requested Fullscreen Mode Is Not Supported By\nYour Video Card. Use Windowed Mode Instead?","Noobtracer",MB_YESNO|MB_ICONEXCLAMATION)==IDYES)
			{
				fullscreen=FALSE;		// Windowed Mode Selected.  Fullscreen = FALSE
			}
			else
			{
				// Pop Up A Message Box Letting User Know The Program Is Closing.
				MessageBox(NULL,"Program Will Now Close.","ERROR",MB_OK|MB_ICONSTOP);
				return FALSE;									// Return FALSE
			}
		}
	}

	if (fullscreen)												// Are We Still In Fullscreen Mode?
	{
		dwExStyle=WS_EX_APPWINDOW;								// Window Extended Style
		dwStyle=WS_POPUP;										// Windows Style
		ShowCursor(FALSE);										// Hide Mouse Pointer
	}
	else
	{
		dwExStyle=WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;			// Window Extended Style
		dwStyle=WS_OVERLAPPEDWINDOW;							// Windows Style
	}

	AdjustWindowRectEx(&WindowRect, dwStyle, FALSE, dwExStyle);		// Adjust Window To True Requested Size

	// Create The Window
	if (!(hWnd=CreateWindowEx(	dwExStyle,							// Extended Style For The Window
								"OpenGL",							// Class Name
								title,								// Window Title
								dwStyle |							// Defined Window Style
								WS_CLIPSIBLINGS |					// Required Window Style
								WS_CLIPCHILDREN,					// Required Window Style
								0, 0,								// Window Position
								WindowRect.right-WindowRect.left,	// Calculate Window Width
								WindowRect.bottom-WindowRect.top,	// Calculate Window Height
								NULL,								// No Parent Window
								NULL,								// No Menu
								hInstance,							// Instance
								NULL)))								// Dont Pass Anything To WM_CREATE
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Window Creation Error.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	static	PIXELFORMATDESCRIPTOR pfd=				// pfd Tells Windows How We Want Things To Be
	{
		sizeof(PIXELFORMATDESCRIPTOR),				// Size Of This Pixel Format Descriptor
		1,											// Version Number
		PFD_DRAW_TO_WINDOW |						// Format Must Support Window
		PFD_SUPPORT_OPENGL |						// Format Must Support OpenGL
		PFD_DOUBLEBUFFER,							// Must Support Double Buffering
		PFD_TYPE_RGBA,								// Request An RGBA Format
		bits,										// Select Our Color Depth
		0, 0, 0, 0, 0, 0,							// Color Bits Ignored
		0,											// No Alpha Buffer
		0,											// Shift Bit Ignored
		0,											// No Accumulation Buffer
		0, 0, 0, 0,									// Accumulation Bits Ignored
		16,											// 16Bit Z-Buffer (Depth Buffer)
		0,											// No Stencil Buffer
		0,											// No Auxiliary Buffer
		PFD_MAIN_PLANE,								// Main Drawing Layer
		0,											// Reserved
		0, 0, 0										// Layer Masks Ignored
	};

	if (!(hDC=GetDC(hWnd)))							// Did We Get A Device Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Device Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if (!(PixelFormat=ChoosePixelFormat(hDC,&pfd)))	// Did Windows Find A Matching Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!SetPixelFormat(hDC,PixelFormat,&pfd))		// Are We Able To Set The Pixel Format?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if (!(hRC=wglCreateContext(hDC)))				// Are We Able To Get A Rendering Context?
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	if(!wglMakeCurrent(hDC,hRC))					// Try To Activate The Rendering Context
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	ShowWindow(hWnd,SW_SHOW);						// Show The Window
	SetForegroundWindow(hWnd);						// Slightly Higher Priority
	SetFocus(hWnd);									// Sets Keyboard Focus To The Window

	/*if(!SetTimer(hWnd, INPUT_TIMER_ID, 100, (TIMERPROC)MyTimerProc))
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Can't Activate Input Timer.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}*/

	ReSizeGLScene(width, height);					// Set Up Our Perspective GL Screen

	if (!InitGL())									// Initialize Our Newly Created GL Window
	{
		KillGLWindow();								// Reset The Display
		MessageBox(NULL,"Initialization Failed.","ERROR",MB_OK|MB_ICONEXCLAMATION);
		return FALSE;								// Return FALSE
	}

	return TRUE;									// Success
}

LRESULT CALLBACK WndProc(	HWND	hWnd,			// Handle For This Window
							UINT	uMsg,			// Message For This Window
							WPARAM	wParam,			// Additional Message Information
							LPARAM	lParam)			// Additional Message Information
{
	switch (uMsg)									// Check For Windows Messages
	{
		case WM_ACTIVATE:							// Watch For Window Activate Message
		{
			if (!HIWORD(wParam))					// Check Minimization State
			{
				active=TRUE;						// Program Is Active
			}
			else
			{
				active=FALSE;						// Program Is No Longer Active
			}

			return 0;								// Return To The Message Loop
		}

		case WM_SYSCOMMAND:
		{
			switch (wParam)
			{
				case SC_SCREENSAVE:
				case SC_MONITORPOWER:
					return 0;
			}
			break;
		}

		case WM_CLOSE:								// Did We Receive A Close Message?
		{
			PostQuitMessage(0);						// Send A Quit Message
			return 0;								// Jump Back
		}

		/*case WM_RBUTTONDOWN:
		{
			if (!capturing)
			{
				capturing = true;
				SetCapture(hWnd);
				//ShowCursor(false);
				coord.x = GET_X_LPARAM(lParam);
		    	coord.y = GET_Y_LPARAM(lParam);
			}

			return 0;
		}

		case WM_RBUTTONUP:
		{
			if (capturing)
			{
				capturing = false;
				ReleaseCapture();
				//ShowCursor(true);
			}
			return 0;
		}

		case WM_MOUSEMOVE:
		{
			int x = GET_X_LPARAM(lParam);
			int y = GET_Y_LPARAM(lParam);
			if (capturing && coord.x != x && coord.y != y)
			{
			    mouseDelta.x += coord.x - GET_X_LPARAM(lParam);
			    mouseDelta.y += coord.y - GET_Y_LPARAM(lParam);
				SetCursorPos(coord.x, coord.y);
			}
			return 0;
		}*/

		case WM_KEYDOWN:							// Is A Key Being Held Down?
		{
			keys[wParam] = TRUE;					// If So, Mark It As TRUE
			switch(wParam)
			{
				case VK_F1:
					Settings::ShadowTracing = !Settings::ShadowTracing;
					break;
				case VK_F2:
					if (Settings::LightingModel == Settings::LightingModelType::Flat)
						Settings::LightingModel = Settings::LightingModelType::Lambert;
					else if (Settings::LightingModel == Settings::LightingModelType::Lambert)
						Settings::LightingModel = Settings::LightingModelType::Phong;
					else if (Settings::LightingModel == Settings::LightingModelType::Phong)
						Settings::LightingModel = Settings::LightingModelType::Flat;
					break;
				case VK_F4:
					Settings::PathTracing = !Settings::PathTracing;
					break;
				case VK_F5:
					Settings::PathTracingSamples = Settings::PathTracingSamples / 2;
					break;
				case VK_F6:
					Settings::PathTracingSamples = max(Settings::PathTracingSamples * 2, 1);
					break;
			}
			
			return 0;								// Jump Back

		}

		case WM_KEYUP:								// Has A Key Been Released?
		{
			keys[wParam] = FALSE;					// If So, Mark It As FALSE
			return 0;								// Jump Back
		}

		case WM_SIZE:								// Resize The OpenGL Window
		{
			ReSizeGLScene(LOWORD(lParam),HIWORD(lParam));  // LoWord=Width, HiWord=Height
			return 0;								// Jump Back
		}
	}

	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}


int argc = 0;
char** argv = NULL;

int WINAPI WinMain(	HINSTANCE	hInstance,			// Instance
					HINSTANCE	hPrevInstance,		// Previous Instance
					LPSTR		lpCmdLine,			// Command Line Parameters
					int			nCmdShow)			// Window Show State
{
	argc = __argc;
	argv = __argv;
	if (argc > 1)
	{
		char* param_1 = argv[1];
		CreateScene(param_1);
	}
	else
	{
		//TestLoading();
		CreateScene();
	}
	MSG		msg;									// Windows Message Structure
	BOOL	done=FALSE;								// Bool Variable To Exit Loop

	// Ask The User Which Screen Mode They Prefer
	/*if (MessageBox(NULL,"Would You Like To Run In Fullscreen Mode?", "Start FullScreen?",MB_YESNO|MB_ICONQUESTION)==IDNO)
	{
		fullscreen=FALSE;							// Windowed Mode
	}*/
	fullscreen=FALSE;							// Windowed Mode
	// Create Our OpenGL Window
	if (!CreateGLWindow("raytracer",768,480,32,fullscreen))
	{
		return 0;									// Quit If Window Was Not Created
	}

	while(!done)									// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))	// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)				// Have We Received A Quit Message?
			{
				done=TRUE;							// If So done=TRUE
			}
			else									// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);				// Translate The Message
				DispatchMessage(&msg);				// Dispatch The Message
			}
		}
		else										// If There Are No Messages
		{
			// Draw The Scene.  Watch For ESC Key And Quit Messages From DrawGLScene()
			if ((active && !DrawGLScene()) || keys[VK_ESCAPE])	// Active?  Was There A Quit Received?
			{
				done=TRUE;							// ESC or DrawGLScene Signalled A Quit
			}
			else									// Not Time To Quit, Update Screen
			{
				SwapBuffers(hDC);					// Swap Buffers (Double Buffering)
			}

			/*if (keys[VK_F1])						// Is F1 Being Pressed?
			{
				keys[VK_F1]=FALSE;					// If So Make Key FALSE
				KillGLWindow();						// Kill Our Current Window
				fullscreen=!fullscreen;				// Toggle Fullscreen / Windowed Mode
				// Recreate Our OpenGL Window
				if (!CreateGLWindow("unnamed realtime raytracer",320,240,32,fullscreen))
				{
					return 0;						// Quit If Window Was Not Created
				}
			}*/
		}
	}

	// Shutdown
	KillGLWindow();									// Kill The Window
	myfont.Destroy();
	delete scene;
	//delete tree;
	delete tree2;
	//myStopWatch.stopTimer();

	return (msg.wParam);							// Exit The Program
}

